@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     flask_security.forms
     ~~~~~~~~~~~~~~~~~~~~
@@ -7,7 +6,7 @@
 
     :copyright: (c) 2012 by Matt Wright.
     :copyright: (c) 2017 by CERN.
-    :copyright: (c) 2019 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2020 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
@@ -16,7 +15,7 @@ import inspect
 from flask import Markup, current_app, request
 from flask_login import current_user
 from flask_wtf import FlaskForm as BaseForm
-from speaklater import is_lazy_string, make_lazy_string
+from werkzeug.local import LocalProxy
 from wtforms import (
     BooleanField,
     Field,
@@ -28,26 +27,34 @@ from wtforms import (
     ValidationError,
     validators,
 )
+from wtforms.fields.html5 import EmailField
+from wtforms.validators import StopValidation
 
+from .babel import is_lazy_string, make_lazy_string
 from .confirmable import requires_confirmation
 from .utils import (
     _,
     _datastore,
     config_value,
     do_flash,
+    find_user,
+    get_identity_attribute,
     get_message,
     hash_password,
     localize_callback,
     url_for_security,
     validate_redirect_url,
 )
-from .twofactor import verify_totp
+
+# Convenient references
+_security = LocalProxy(lambda: current_app.extensions["security"])
 
 _default_field_labels = {
     "email": _("Email Address"),
     "password": _("Password"),
     "remember_me": _("Remember Me"),
     "login": _("Login"),
+    "signin": _("Sign In"),
     "register": _("Register"),
     "send_confirmation": _("Resend Confirmation Instructions"),
     "recover_password": _("Recover Password"),
@@ -60,10 +67,16 @@ _default_field_labels = {
     "change_method": _("Change Method"),
     "phone": _("Phone Number"),
     "code": _("Authentication Code"),
+    "submit": _("Submit"),
+    "submitcode": _("Submit Code"),
+    "error": _("Error(s)"),
+    "identity": _("Identity"),
+    "sendcode": _("Send Code"),
+    "passcode": _("Passcode"),
 }
 
 
-class ValidatorMixin(object):
+class ValidatorMixin:
     """
     This is called at import time - so there is no app context.
     Validators have state - namely self.message - but we need that
@@ -79,7 +92,7 @@ class ValidatorMixin(object):
             del kwargs["message"]
         else:
             self._original_message = None
-        super(ValidatorMixin, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def __call__(self, form, field):
         if self._original_message and (
@@ -91,7 +104,7 @@ class ValidatorMixin(object):
                 self.message = make_lazy_string(_local_xlate, cv[0])
             else:
                 self.message = self._original_message
-        return super(ValidatorMixin, self).__call__(form, field)
+        return super().__call__(form, field)
 
 
 class EqualTo(ValidatorMixin, validators.EqualTo):
@@ -102,29 +115,42 @@ class Required(ValidatorMixin, validators.DataRequired):
     pass
 
 
-class Email(ValidatorMixin, validators.Email):
-    pass
-
-
 class Length(ValidatorMixin, validators.Length):
     pass
 
 
+class EmailValidation:
+    """Simple interface to email_validator."""
+
+    def __call__(self, form, field):
+        if field.data is None:  # pragma: no cover
+            raise ValidationError(get_message("EMAIL_NOT_PROVIDED")[0])
+
+        try:
+            field.data = _security._mail_util.validate(field.data)
+        except ValueError:
+            msg = get_message("INVALID_EMAIL_ADDRESS")[0]
+            # we stop further validators if email isn't valid.
+            # TODO: email_validator provides some really nice error messages - however
+            # they aren't localized. And there isn't an easy way to add multiple
+            # errors at once.
+            raise StopValidation(msg)
+
+
 email_required = Required(message="EMAIL_NOT_PROVIDED")
-email_validator = Email(message="INVALID_EMAIL_ADDRESS")
+email_validator = EmailValidation()
 password_required = Required(message="PASSWORD_NOT_PROVIDED")
-password_length = Length(min=6, max=128, message="PASSWORD_INVALID_LENGTH")
 
 
 def _local_xlate(text):
-    """ LazyStrings need to be evaluated in the context of a request
+    """LazyStrings need to be evaluated in the context of a request
     where _security.i18_domain is available.
     """
     return localize_callback(text)
 
 
 def get_form_field_label(key):
-    """ This is called during import since form fields are declared as part of
+    """This is called during import since form fields are declared as part of
     class. Thus can't call 'localize_callback' until we need to actually
     translate/render form.
     """
@@ -132,13 +158,52 @@ def get_form_field_label(key):
 
 
 def unique_user_email(form, field):
-    if _datastore.get_user(field.data) is not None:
-        msg = get_message("EMAIL_ALREADY_ASSOCIATED", email=field.data)[0]
+    uia_email = get_identity_attribute("email")
+    norm_email = _security._mail_util.normalize(field.data)
+    if (
+        _datastore.find_user(
+            case_insensitive=uia_email.get("case_insensitive", False), email=norm_email
+        )
+        is not None
+    ):
+        msg = get_message("EMAIL_ALREADY_ASSOCIATED", email=norm_email)[0]
         raise ValidationError(msg)
 
 
+def unique_identity_attribute(form, field):
+    """A validator that checks the field data against all configured
+    SECURITY_USER_IDENTITY_ATTRIBUTES.
+    This can be used as part of registration.
+
+    Be aware that the "mapper" function likely also normalizes the input in addition
+    to validating it.
+
+    :param form:
+    :param field:
+    :return: Nothing; if field data corresponds to an existing User, ValidationError
+        is raised.
+    """
+    for mapping in config_value("USER_IDENTITY_ATTRIBUTES"):
+        attr = list(mapping.keys())[0]
+        details = mapping[attr]
+        idata = details["mapper"](field.data)
+        if idata:
+            if _datastore.find_user(
+                case_insensitive=details.get("case_insensitive", False), **{attr: idata}
+            ):
+                msg = get_message(
+                    "IDENTITY_ALREADY_ASSOCIATED", attr=attr, value=idata
+                )[0]
+                raise ValidationError(msg)
+
+
 def valid_user_email(form, field):
-    form.user = _datastore.get_user(field.data)
+    # Verify email exists in DB - be sure to normalize first.
+    uia_email = get_identity_attribute("email")
+    norm_email = _security._mail_util.normalize(field.data)
+    form.user = _datastore.find_user(
+        case_insensitive=uia_email.get("case_insensitive", False), email=norm_email
+    )
     if form.user is None:
         raise ValidationError(get_message("USER_DOES_NOT_EXIST")[0])
 
@@ -147,25 +212,25 @@ class Form(BaseForm):
     def __init__(self, *args, **kwargs):
         if current_app.testing:
             self.TIME_LIMIT = None
-        super(Form, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
 
 class EmailFormMixin:
-    email = StringField(
+    email = EmailField(
         get_form_field_label("email"), validators=[email_required, email_validator]
     )
 
 
 class UserEmailFormMixin:
     user = None
-    email = StringField(
+    email = EmailField(
         get_form_field_label("email"),
         validators=[email_required, email_validator, valid_user_email],
     )
 
 
 class UniqueEmailFormMixin:
-    email = StringField(
+    email = EmailField(
         get_form_field_label("email"),
         validators=[email_required, email_validator, unique_user_email],
     )
@@ -179,8 +244,7 @@ class PasswordFormMixin:
 
 class NewPasswordFormMixin:
     password = PasswordField(
-        get_form_field_label("password"),
-        validators=[password_required, password_length],
+        get_form_field_label("password"), validators=[password_required]
     )
 
 
@@ -207,14 +271,27 @@ class NextFormMixin:
 class RegisterFormMixin:
     submit = SubmitField(get_form_field_label("register"))
 
-    def to_dict(form):
-        def is_field_and_user_attr(member):
-            return isinstance(member, Field) and hasattr(
-                _datastore.user_model, member.name
-            )
+    def to_dict(self, only_user):
+        """
+        Return form data as dictionary
+        :param only_user: bool, if True then only fields that have
+        corresponding members in UserModel are returned
+        :return: dict
+        """
 
-        fields = inspect.getmembers(form, is_field_and_user_attr)
-        return dict((key, value.data) for key, value in fields)
+        def is_field_and_user_attr(member):
+            if not isinstance(member, Field):
+                return False
+
+            # If only fields recorded on UserModel should be returned,
+            # perform check on user model, else return True
+            if only_user is True:
+                return hasattr(_datastore.user_model, member.name)
+            else:
+                return True
+
+        fields = inspect.getmembers(self, is_field_and_user_attr)
+        return {key: value.data for key, value in fields}
 
 
 class SendConfirmationForm(Form, UserEmailFormMixin):
@@ -223,12 +300,12 @@ class SendConfirmationForm(Form, UserEmailFormMixin):
     submit = SubmitField(get_form_field_label("send_confirmation"))
 
     def __init__(self, *args, **kwargs):
-        super(SendConfirmationForm, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         if request.method == "GET":
             self.email.data = request.args.get("email", None)
 
     def validate(self):
-        if not super(SendConfirmationForm, self).validate():
+        if not super().validate():
             return False
         if self.user.confirmed_at is not None:
             self.email.errors.append(get_message("ALREADY_CONFIRMED")[0])
@@ -241,10 +318,18 @@ class ForgotPasswordForm(Form, UserEmailFormMixin):
 
     submit = SubmitField(get_form_field_label("recover_password"))
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.requires_confirmation = False
+
     def validate(self):
-        if not super(ForgotPasswordForm, self).validate():
+        if not super().validate():
             return False
-        if requires_confirmation(self.user):
+        if not self.user.is_active:
+            self.email.errors.append(get_message("DISABLED_ACCOUNT")[0])
+            return False
+        self.requires_confirmation = requires_confirmation(self.user)
+        if self.requires_confirmation:
             self.email.errors.append(get_message("CONFIRMATION_REQUIRED")[0])
             return False
         return True
@@ -256,10 +341,10 @@ class PasswordlessLoginForm(Form, UserEmailFormMixin):
     submit = SubmitField(get_form_field_label("send_login_link"))
 
     def __init__(self, *args, **kwargs):
-        super(PasswordlessLoginForm, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def validate(self):
-        if not super(PasswordlessLoginForm, self).validate():
+        if not super().validate():
             return False
         if not self.user.is_active:
             self.email.errors.append(get_message("DISABLED_ACCOUNT")[0])
@@ -270,7 +355,7 @@ class PasswordlessLoginForm(Form, UserEmailFormMixin):
 class LoginForm(Form, NextFormMixin):
     """The default login form"""
 
-    email = StringField(get_form_field_label("email"), validators=[email_required])
+    email = EmailField(get_form_field_label("email"), validators=[email_required])
     password = PasswordField(
         get_form_field_label("password"), validators=[password_required]
     )
@@ -278,7 +363,7 @@ class LoginForm(Form, NextFormMixin):
     submit = SubmitField(get_form_field_label("login"))
 
     def __init__(self, *args, **kwargs):
-        super(LoginForm, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         if not self.next.data:
             self.next.data = request.args.get("next", "")
         self.remember.default = config_value("DEFAULT_REMEMBER_ME")
@@ -287,18 +372,22 @@ class LoginForm(Form, NextFormMixin):
             and not self.password.description
         ):
             html = Markup(
-                u'<a href="{url}">{message}</a>'.format(
+                '<a href="{url}">{message}</a>'.format(
                     url=url_for_security("forgot_password"),
                     message=get_message("FORGOT_PASSWORD")[0],
                 )
             )
             self.password.description = html
+        self.requires_confirmation = False
 
     def validate(self):
-        if not super(LoginForm, self).validate():
+        if not super().validate():
             return False
 
-        self.user = _datastore.get_user(self.email.data)
+        # Historically, this used get_user() which would look at all
+        # USER_IDENTITY_ATTRIBUTES - even though the field name is 'email'
+        # We keep that behavior (for now) as we transition to find_user.
+        self.user = find_user(self.email.data)
 
         if self.user is None:
             self.email.errors.append(get_message("USER_DOES_NOT_EXIST")[0])
@@ -310,10 +399,12 @@ class LoginForm(Form, NextFormMixin):
             # Reduce timing variation between existing and non-existing users
             hash_password(self.password.data)
             return False
+        self.password.data = _security._password_util.normalize(self.password.data)
         if not self.user.verify_and_update_password(self.password.data):
             self.password.errors.append(get_message("INVALID_PASSWORD")[0])
             return False
-        if requires_confirmation(self.user):
+        self.requires_confirmation = requires_confirmation(self.user)
+        if self.requires_confirmation:
             self.email.errors.append(get_message("CONFIRMATION_REQUIRED")[0])
             return False
         if not self.user.is_active:
@@ -322,15 +413,92 @@ class LoginForm(Form, NextFormMixin):
         return True
 
 
-class ConfirmRegisterForm(
-    Form, RegisterFormMixin, UniqueEmailFormMixin, NewPasswordFormMixin
-):
-    pass
+class VerifyForm(Form, PasswordFormMixin):
+    """The verify authentication form"""
+
+    user = None
+    submit = SubmitField(get_form_field_label("verify_password"))
+
+    def validate(self):
+        if not super().validate():
+            return False
+
+        self.user = current_user
+        self.password.data = _security._password_util.normalize(self.password.data)
+        if not self.user.verify_and_update_password(self.password.data):
+            self.password.errors.append(get_message("INVALID_PASSWORD")[0])
+            return False
+        return True
 
 
-class RegisterForm(ConfirmRegisterForm, PasswordConfirmFormMixin, NextFormMixin):
+class ConfirmRegisterForm(Form, RegisterFormMixin, UniqueEmailFormMixin):
+    """This form is used for registering when 'confirmable' is set.
+    The only difference between this and the other RegisterForm is that
+    this one doesn't require re-typing in the password...
+    """
+
+    # Password optional when Unified Signin enabled.
+    password = PasswordField(
+        get_form_field_label("password"), validators=[validators.Optional()]
+    )
+
+    def validate(self):
+        if not super().validate():
+            return False
+
+        # To support unified sign in - we permit registering with no password.
+        if not config_value("UNIFIED_SIGNIN"):
+            # password required
+            if not self.password.data or not self.password.data.strip():
+                self.password.errors.append(get_message("PASSWORD_NOT_PROVIDED")[0])
+                return False
+
+        if self.password.data:
+            # We do explicit validation here for passwords
+            # (rather than write a validator class) for 2 reasons:
+            # 1) We want to control which fields are passed -
+            #    sometimes that's current_user
+            #    other times it's the registration fields.
+            # 2) We want to be able to return multiple error messages.
+            rfields = {}
+            for k, v in self.data.items():
+                if hasattr(_datastore.user_model, k):
+                    rfields[k] = v
+            del rfields["password"]
+            pbad, self.password.data = _security._password_util.validate(
+                self.password.data, True, **rfields
+            )
+            if pbad:
+                self.password.errors.extend(pbad)
+                return False
+        return True
+
+
+class RegisterForm(ConfirmRegisterForm, NextFormMixin):
+
+    # Password optional when Unified Signin enabled.
+    password_confirm = PasswordField(
+        get_form_field_label("retype_password"),
+        validators=[
+            EqualTo("password", message="RETYPE_PASSWORD_MISMATCH"),
+            validators.Optional(),
+        ],
+    )
+
+    def validate(self):
+        if not super().validate():
+            return False
+        if not config_value("UNIFIED_SIGNIN"):
+            # password_confirm required
+            if not self.password_confirm.data or not self.password_confirm.data.strip():
+                self.password_confirm.errors.append(
+                    get_message("PASSWORD_NOT_PROVIDED")[0]
+                )
+                return False
+        return True
+
     def __init__(self, *args, **kwargs):
-        super(RegisterForm, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         if not self.next.data:
             self.next.data = request.args.get("next", "")
 
@@ -340,13 +508,24 @@ class ResetPasswordForm(Form, NewPasswordFormMixin, PasswordConfirmFormMixin):
 
     submit = SubmitField(get_form_field_label("reset_password"))
 
+    def validate(self):
+        if not super().validate():
+            return False
+
+        pbad, self.password.data = _security._password_util.validate(
+            self.password.data, False, user=current_user
+        )
+        if pbad:
+            self.password.errors.extend(pbad)
+            return False
+        return True
+
 
 class ChangePasswordForm(Form, PasswordFormMixin):
     """The default change password form"""
 
     new_password = PasswordField(
-        get_form_field_label("new_password"),
-        validators=[password_required, password_length],
+        get_form_field_label("new_password"), validators=[password_required]
     )
 
     new_password_confirm = PasswordField(
@@ -360,15 +539,22 @@ class ChangePasswordForm(Form, PasswordFormMixin):
     submit = SubmitField(get_form_field_label("change_password"))
 
     def validate(self):
-        if not super(ChangePasswordForm, self).validate():
+        if not super().validate():
             return False
 
+        self.password.data = _security._password_util.normalize(self.password.data)
         if not current_user.verify_and_update_password(self.password.data):
             self.password.errors.append(get_message("INVALID_PASSWORD")[0])
             return False
         if self.password.data == self.new_password.data:
             self.password.errors.append(get_message("PASSWORD_IS_THE_SAME")[0])
             return False
+        pbad, self.new_password.data = _security._password_util.validate(
+            self.new_password.data, False, user=current_user
+        )
+        if pbad:
+            self.new_password.errors.extend(pbad)
+            return False
         return True
 
 
@@ -378,9 +564,12 @@ class TwoFactorSetupForm(Form, UserEmailFormMixin):
     setup = RadioField(
         "Available Methods",
         choices=[
-            ("mail", "Set Up Using Mail"),
-            ("google_authenticator", "Set Up Using Google Authenticator"),
-            ("sms", "Set Up Using SMS"),
+            ("email", "Set up using email"),
+            (
+                "authenticator",
+                "Set up using an authenticator app (e.g. google, lastpass, authy)",
+            ),
+            ("sms", "Set up using SMS"),
             ("disable", "Disable two factor authentication"),
         ],
     )
@@ -388,16 +577,31 @@ class TwoFactorSetupForm(Form, UserEmailFormMixin):
     submit = SubmitField(get_form_field_label("submit"))
 
     def __init__(self, *args, **kwargs):
-        super(TwoFactorSetupForm, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def validate(self):
+        # TODO: the super class validate is never called - thus we have to
+        # initialize errors to lists below. It also means that 'email' is never
+        # validated - though it isn't required so the mixin might not be correct.
         choices = config_value("TWO_FACTOR_ENABLED_METHODS")
+        if "email" in choices:
+            # backwards compat
+            choices.append("mail")
         if not config_value("TWO_FACTOR_REQUIRED"):
             choices.append("disable")
         if "setup" not in self.data or self.data["setup"] not in choices:
             self.setup.errors = list()
             self.setup.errors.append(get_message("TWO_FACTOR_METHOD_NOT_AVAILABLE")[0])
             return False
+        if self.setup.data == "sms" and len(self.phone.data) > 0:
+            # Somewhat bizarre - but this isn't required the first time around
+            # when they select "sms". Then they get a field to fill out with
+            # phone number, then Submit again.
+            msg = _security._phone_util.validate_phone_number(self.phone.data)
+            if msg:
+                self.phone.errors = list()
+                self.phone.errors.append(msg)
+                return False
 
         return True
 
@@ -406,16 +610,19 @@ class TwoFactorVerifyCodeForm(Form, UserEmailFormMixin):
     """The Two-factor token validation form"""
 
     code = StringField(get_form_field_label("code"))
-    submit = SubmitField(get_form_field_label("submit code"))
+    submit = SubmitField(get_form_field_label("submitcode"))
 
     def __init__(self, *args, **kwargs):
-        super(TwoFactorVerifyCodeForm, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def validate(self):
         # codes sent by sms or mail will be valid for another window cycle
-        if self.primary_method == "google_authenticator":
-            self.window = config_value("TWO_FACTOR_GOOGLE_AUTH_VALIDITY")
-        elif self.primary_method == "mail":
+        if (
+            self.primary_method == "google_authenticator"
+            or self.primary_method == "authenticator"
+        ):
+            self.window = config_value("TWO_FACTOR_AUTHENTICATOR_VALIDITY")
+        elif self.primary_method == "email" or self.primary_method == "mail":
             self.window = config_value("TWO_FACTOR_MAIL_VALIDITY")
         elif self.primary_method == "sms":
             self.window = config_value("TWO_FACTOR_SMS_VALIDITY")
@@ -423,9 +630,10 @@ class TwoFactorVerifyCodeForm(Form, UserEmailFormMixin):
             return False
 
         # verify entered token with user's totp secret
-        if not verify_totp(
+        if not _security._totp_factory.verify_totp(
             token=self.code.data,
-            totp_secret=self.user.tf_totp_secret,
+            totp_secret=self.tf_totp_secret,
+            user=self.user,
             window=self.window,
         ):
             self.code.errors = list()
@@ -436,24 +644,7 @@ class TwoFactorVerifyCodeForm(Form, UserEmailFormMixin):
         return True
 
 
-class TwoFactorVerifyPasswordForm(Form, PasswordFormMixin):
-    """The verify password form"""
-
-    submit = SubmitField(get_form_field_label("verify_password"))
-
-    def validate(self):
-        if not super(TwoFactorVerifyPasswordForm, self).validate():
-            return False
-
-        self.user = current_user
-        if not self.user.verify_and_update_password(self.password.data):
-            self.password.errors.append(get_message("INVALID_PASSWORD")[0])
-            return False
-
-        return True
-
-
-class TwoFactorRescueForm(Form, UserEmailFormMixin):
+class TwoFactorRescueForm(Form):
     """The Two-factor Rescue validation form """
 
     help_setup = RadioField(
@@ -466,7 +657,9 @@ class TwoFactorRescueForm(Form, UserEmailFormMixin):
     submit = SubmitField(get_form_field_label("submit"))
 
     def __init__(self, *args, **kwargs):
-        super(TwoFactorRescueForm, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def validate(self):
+        if not super().validate():
+            return False
         return True
